"""Python Revenue Assurance Process Optimizer."""

from .main.scheduler import Scheduler
from .main.control import Control
from .web import WEBAPI


__author__ = 'Timur Faradzhov'
__copyright__ = 'Copyright 2020, The RAPO project'
__credits__ = ['Timur Faradzhov']

__license__ = 'MIT'
__version__ = '0.3.2'
__maintainer__ = 'Timur Faradzhov'
__email__ = 'timurfaradzhov@gmail.com'
__status__ = 'Development'

__all__ = [Scheduler, Control, WEBAPI]



"""Contains web API elements."""

import datetime as dt
import getpass
import os
import platform
import signal
import subprocess as sp
import sys

import psutil

from .api import app
from ..config import config
from ..database import db
from ..reader import reader


class WEBAPI():
    """Represents RAPO web API."""

    def __init__(self, host=None, port=None, dev=False):
        argv = sys.argv[1:].copy()

        self.app = app
        self.host = host or config['API'].get('host') or '127.0.0.1'
        self.port = port or config['API'].get('port') or 8080
        self.dev = True if dev is True or 'dev' in argv else False

        self.table = db.tables.web_api
        self.record = reader.read_web_api_record()
        if self.record and self.record['status'] == 'Y':
            self.server = self.record['server']
            self.username = self.record['username']
            self.pid = int(self.record['pid'])
            self.start_date = self.record['start_date']
            self.stop_date = self.record['stop_date']
            self.status = True if self.record['status'] == 'Y' else False
        else:
            self.server = platform.node()
            self.username = getpass.getuser()
            self.pid = None
            self.start_date = None
            self.stop_date = None
            self.status = None

        if argv:
            if argv[0] == 'start':
                self.start()
            elif argv[0] == 'stop':
                self.stop()
        pass

    def start(self):
        """Start web API server."""
        conn = db.connect()
        if self.status is True and self.pid and psutil.pid_exists(self.pid):
            message = f'web API already running at PID {self.pid}'
            raise Exception(message)
        app = f'{self.app.name}:app'
        env = os.environ.copy()
        self.start_date = dt.datetime.now()
        self.status = True
        if self.dev is True:
            script = ['flask', 'run']
            args = ['--host', self.host, '--port', str(self.port)]
            cmd = [arg for arg in [*script, *args] if arg is not None]
            env['FLASK_APP'] = app
            env['FLASK_ENV'] = 'development'
            try:
                proc = sp.Popen(cmd, env=env)
                self.pid = proc.pid
                update = (self.table.update()
                                    .values(server=self.server,
                                            username=self.username,
                                            pid=self.pid,
                                            url=f'{self.host}:{self.port}',
                                            debug='X',
                                            start_date=self.start_date,
                                            stop_date=self.stop_date,
                                            status='Y'))
                conn.execute(update)
                conn.close()
                proc.wait()
            except KeyboardInterrupt:
                conn = db.connect()
                self.stop_date = dt.datetime.now()
                update = (self.table.update()
                                    .values(stop_date=self.stop_date,
                                            status='N'))
                conn.execute(update)
                conn.close()
                proc.terminate()
        else:
            script = 'waitress-serve'
            args = ['--host', self.host, '--port', str(self.port), app]
            cmd = [arg for arg in [script, *args] if arg is not None]
            proc = sp.Popen(cmd, env=env, stdout=sp.DEVNULL, stderr=sp.DEVNULL)
            self.pid = proc.pid
            update = self.table.update().values(server=self.server,
                                                username=self.username,
                                                pid=self.pid,
                                                url=f'{self.host}:{self.port}',
                                                debug=None,
                                                start_date=self.start_date,
                                                stop_date=self.stop_date,
                                                status='Y')
            conn.execute(update)
            conn.close()
        pass

    def stop(self):
        """Stop web API server."""
        conn = db.connect()
        if self.status is True:
            self.stop_date = dt.datetime.now()
            self.status = False
            if psutil.pid_exists(self.pid):
                os.kill(self.pid, signal.SIGTERM)
            update = self.table.update().values(stop_date=self.stop_date,
                                                status='N')
            conn.execute(update)
        pass

    pass



"""Contains web API application and routes."""

import os

import flask

from .auth import auth
from .response import OK

from ...reader import reader

from ...main.control import Control


app = flask.Flask(__name__)


@app.route('/api/help')
@auth.login_required
def help():
    """Get help message."""
    path = os.path.join(os.path.dirname(__file__), 'templates/help.html')
    lines = open(path, 'r').readlines()
    text = ''.join(lines)
    return text


@app.route('/api/run-control', methods=['POST'])
@auth.login_required
def run_control():
    """Run control and get its result in JSON."""
    request = flask.request
    name = request.args['name']
    date = request.args.get('date')
    date_from = request.args.get('date_from')
    date_to = request.args.get('date_to')
    control = Control(name, date_from=date_from, date_to=date_to, date=date)
    control.process()
    response = flask.jsonify(name=control.name,
                             date_from=control.date_from,
                             date_to=control.date_to,
                             process_id=control.process_id,
                             start_date=control.start_date,
                             end_date=control.end_date,
                             status=control.status,
                             fetched=control.fetched,
                             success=control.success,
                             errors=control.errors,
                             error_level=control.error_level,
                             fetched_a=control.fetched_a,
                             fetched_b=control.fetched_b,
                             success_a=control.success_a,
                             success_b=control.success_b,
                             errors_a=control.errors_a,
                             errors_b=control.errors_b,
                             error_level_a=control.error_level_a,
                             error_level_b=control.error_level_b)
    return response


@app.route('/api/cancel-control', methods=['POST'])
@auth.login_required
def cancel_control():
    """Cancel running control."""
    request = flask.request
    process_id = int(request.args['id'])
    control = Control(process_id=process_id)
    control.cancel()
    return OK


@app.route('/api/get-running-controls')
@auth.login_required
def get_running_controls():
    """Get list of currently running controls in JSON."""
    rows = reader.read_running_controls()
    response = flask.jsonify(rows)
    return response


@app.route('/api/revoke-control-run', methods=['DELETE'])
@auth.login_required
def revoke_control_run():
    """Revoke patricular control run."""
    request = flask.request
    process_id = int(request.args['id'])
    control = Control(process_id=process_id)
    control.revoke()
    return OK



"""Contains application data reader."""

import sqlalchemy as sa

from .database import db


class Reader():
    """Represents application data reader.

    Reader used to extract aplication data from database to user.
    """

    def read_scheduler_record(self):
        """Get scheduler record from DB table.

        Returns
        -------
        record : dict
            Ordinary dictionary with web API information from DB table.
        """
        conn = db.connect()
        table = db.tables.scheduler
        select = table.select()
        result = conn.execute(select).first()
        record = dict(result)
        return record

    def read_web_api_record(self):
        """Get web API record from DB table.

        Returns
        -------
        record : dict
            Ordinary dictionary with web API information from DB table.
        """
        conn = db.connect()
        table = db.tables.web_api
        select = table.select()
        result = conn.execute(select).first()
        record = dict(result)
        return record

    def read_control_name(self, process_id):
        """Get control name using passed process_id.

        Parameters
        ----------
        process_id : int
            Unique process ID used to load record from DB log.

        Returns
        -------
        control_name : str
            Name of the defined control.
        """
        conn = db.connect()
        log = db.tables.log
        config = db.tables.config
        join = log.join(config, log.c.control_id == config.c.control_id)
        select = (sa.select([config.c.control_name]).select_from(join)
                    .where(log.c.process_id == process_id))
        result = conn.execute(select).first()
        control_name = result.control_name
        return control_name

    def read_control_result(self, process_id):
        """Get control result from DB log table.

        Parameters
        ----------
        process_id : int
            Unique process ID used to load record from DB log.

        Returns
        -------
        record : dict
            Ordinary dictionary with control result.
        """
        if process_id:
            conn = db.connect()
            table = db.tables.log
            select = table.select().where(table.c.process_id == process_id)
            result = conn.execute(select).first()
            if result:
                record = dict(result)
                return record
            else:
                message = f'no process with ID {process_id} found'
                raise ValueError(message)
        else:
            return None
        pass

    def read_control_config(self, control_name):
        """Get dictionary with control configuration from DB.

        Parameters
        ----------
        control_name : str
            Unique control name used to load record from DB configuration.

        Returns
        -------
        record : dict
            Ordinary dictionary with control configuration.
        """
        conn = db.connect()
        table = db.tables.config
        select = table.select().where(table.c.control_name == control_name)
        result = conn.execute(select).first()
        if result:
            record = dict(result)
            return record
        else:
            message = f'no control with name {control_name} found'
            raise ValueError(message)
        pass

    def read_running_controls(self):
        """Get list of running controls."""
        conn = db.connect()
        table = db.tables.log
        select = table.select().where(table.c.status == 'P')
        result = conn.execute(select)
        rows = [dict(row) for row in result]
        return rows

    pass


reader = Reader()



"""Contains application logger."""

import pepperoni

from .config import config


logger = pepperoni.logger(file=True)
logger.configure(format='{isodate}\t{thread}\t{rectype}\t{message}\n')

if config.has_section('LOGGING'):
    LOGGING = config['LOGGING']
    logger.configure(console=LOGGING.getboolean('console'),
                     file=LOGGING.getboolean('file'),
                     info=LOGGING.getboolean('info'),
                     debug=LOGGING.getboolean('debug'),
                     error=LOGGING.getboolean('error'),
                     warning=LOGGING.getboolean('warning'),
                     critical=LOGGING.getboolean('critical'),
                     maxsize=LOGGING.getint('maxsize'),
                     maxdays=LOGGING.getint('maxdays'))



"""Contains application configurator.

Used to transfer parameters from user to application.
"""

import configparser
import os


path = os.path.abspath(os.path.expanduser('~/.rapo/rapo.ini'))
config = configparser.ConfigParser(allow_no_value=True)
config.read(path)



"""Contains application utils."""

import datetime as dt


class Utils():
    """Represents application utils."""

    def to_date(self, input):
        """Normalize input value to datetime object.

        Parameters
        ----------
        input : str, datetime.datetime, None
            Initial value that should be normalized to datetime object.

        Returns
        -------
        output : datetime.datetime
            Nomalized datatime object.
        """
        if input is None:
            output = dt.datetime.now()
        elif isinstance(input, dt.datetime):
            output = input
        elif isinstance(input, str):
            output = dt.datetime.fromisoformat(input)
        elif isinstance(input, (int, float)):
            output = dt.datetime.fromtimestamp(input)
        return output

    def to_lower(self, value):
        """Lower string or return None.

        Parameters
        ----------
        value : str
            Initial value that must be modified.

        Returns
        -------
        value : str or None
            Modified value.
        """
        value = value.lower() if isinstance(value, str) is True else None
        return value


utils = Utils()



"""Contains database instance with application schema."""

import sys

import sqlalchemy as sa
import sqlparse as spe

from .config import config


class Database():
    """Represent database with application schema."""

    ckwargs = {'literal_binds': True}

    def __init__(self):
        vendor = config['DATABASE']['vendor']
        host = config['DATABASE']['host']
        port = config['DATABASE']['port']
        path = config['DATABASE'].get('path')
        sid = config['DATABASE'].get('sid')
        service = config['DATABASE'].get('service')
        user = config['DATABASE']['user']
        password = config['DATABASE']['password']
        if vendor == 'sqlite' and path is not None:
            if sys.platform.startswith('win'):
                url = f'{vendor}:///{path}'
            else:
                url = f'{vendor}:////{path}'
        else:
            url = f'{vendor}://{user}:{password}@{host}:{port}'
            if sid is not None:
                url += f'/{sid}'
            elif service is not None:
                url += f'/?service_name={service}'
        settings = dict(pool_pre_ping=True, max_identifier_length=128)
        self.engine = sa.create_engine(url, **settings)
        self.tables = self.Tables(self)
        self.formatter = self.Formatter(self)
        pass

    class Tables():
        """Represents database tables."""

        def __init__(self, database):
            self.database = database
            self.config = self.load('rapo_config')
            self.log = self.load('rapo_log')
            self.scheduler = self.load('rapo_scheduler')
            self.web_api = self.load('rapo_web_api')
            pass

        def load(self, name):
            """Load one of database table by name.

            Returns
            -------
            table : sqlalchemy.Table
                Database table instance.
            """
            meta = sa.MetaData()
            engine = self.database.engine
            table = sa.Table(name, meta, autoload=True, autoload_with=engine)
            return table

        pass

    class Formatter():
        """Represents database query formatter."""

        def __init__(self, database):
            self.database = database
            pass

        def __call__(self, *queries):
            """Format received raw SQL query into final human-read form."""
            results = []
            for query in queries:
                query = self._prepare(query)
                result = self._parse(query)
                results.append(result)
            strings = '\n'.join(results)
            string = f'{str():->40}\n{strings}\n{str():->40}'
            return string

        def _prepare(self, query):
            query = self._to_string(query)
            query = self._separate(query)
            return query

        def _parse(self, query):
            result = spe.format(query, keyword_case='upper',
                                identifier_case='lower',
                                reindent_aligned=True)
            return result

        def _to_string(self, query):
            if isinstance(query, str):
                return query
            else:
                engine = self.database.engine
                ckwargs = self.database.ckwargs
                query = query.compile(bind=engine, compile_kwargs=ckwargs)
                string = query.string
                return string
            pass

        def _separate(self, query):
            if query.rstrip()[-1] != ';':
                query = f'{query.rstrip()};'
            return query

        pass

    def connect(self):
        """Get database connection.

        Returns
        -------
        connection : sqlalchemy.Connection
            Database connection instance.
        """
        connection = self.engine.connect()
        return connection

    def table(self, name):
        """Get database table.

        Returns
        -------
        table : sqlalchemy.Table
            Database table instance.
        """
        meta = sa.MetaData()
        table = sa.Table(name, meta, autoload=True, autoload_with=self.engine)
        return table

    pass


db = Database()



"""Contains RAPO control interface."""

import datetime as dt
import json
import multiprocessing as mp
import sys
import threading as th
import time as tm
import traceback as tb

import sqlalchemy as sa

from ..database import db
from ..logger import logger
from ..reader import reader
from ..utils import utils


class Control():
    """Represents certain RAPO control and acts like its API.

    Parameters
    ----------
    name : str, optional
        Name of the control from RAPO_CONFIG.
    timestamp : float or None
        Timestamp of this process.
    date_from : str or datetime, optional
        Data source date lower bound.
    date_to : str or datetime, optional
        Data source date upper bound.

    Attributes
    ----------
    name : str
        Name of the control from RAPO_CONFIG.
    timestamp : float or None
        Timestamp of this process.
    parser : rapo.Parser
        Parser instance for this control.
    executor : rapo.Executor
        Executor instance for this control.
    log : sqlalchemy.Table
        RAPO_LOG table object.
    config : sqlalchemy.RowProxy
        RAPO_CONFIG record for this control.
    id : int
        ID of the control from RAPO_CONFIG.
    group : str or None
        Group of the control from RAPO_CONFIG.
    type : str or None
        Type of the control from RAPO_CONFIG. See RAPO_REF_TYPES table.
    subtype: str or None
        Method of the control from RAPO_CONFIG. See RAPO_REF_METHODS table.
    engine : str or None
        Engine of the control from RAPO_CONFIG. See RAPO_REF_ENGINES table.
    process_id : int or None
        Process id from RAPO_LOG.
    pid : int or None
        Process id from RAPO_LOG.
    source_name: str or None
        Data source name.
    source_date_field :
        Data source date field.
    source_name_a : str or None
        Data source A name.
    source_date_field_a :
        Data source A date field.
    source_name_b : str or None
        Data source B name.
    source_date_field_b :
        Data source B date field.
    updated : datetime or None
        Date when process RAPO_LOG record was last updated.
    start_date : datetime or None
        Date when process started.
    end_date : datetime or None
        Date when process finished.
    status : bool or None
        Current process status.
    select : sqlalchemy.Select
        SQL statement to fetch data from data source.
    select_a : sqlalchemy.Select
        SQL statement to fetch data from data source A.
    select_b : sqlalchemy.Select
        SQL statement to fetch data from data source B.
    rule_config : list
        Configuration defining success result.
    error_config : list
        Configuration defining error result.
    output_columns : list or None
        Output column configuration.
    output_columns_a : list or None
        Output A column configuration.
    output_columns_b : list or None
        Output B column configuration.
    need_a : bool or None
        Flag to define whether save output A or not.
    need_b : bool or None
        Flag to define whether save output B or not.
    need_hook : bool or None
        Flag to define whether run database hook procedure or not.
    source_table : sqlalchemy.Table
        Proxy object reflecting data source table.
    source_table_a : sqlalchemy.Table
        Proxy object reflecting data source A table.
    source_table_b : sqlalchemy.Table
        Proxy object reflecting data source B table.
    input_table : sqlalchemy.Table
        Proxy object reflecting table with fetched records from data source.
    input_table_a : sqlalchemy.Table
        Proxy object reflecting table with fetched records from data source A.
    input_table_b : sqlalchemy.Table
        Proxy object reflecting table with fetched records from data source B.
    result_table : sqlalchemy.Table
        Proxy object reflecting table with found matches.
    error_table : sqlalchemy.Table
        Proxy object reflecting table with found discrapancies.
    error_table_a : sqlalchemy.Table
        Proxy object reflecting table with found discrapancies from A.
    error_table_b : sqlalchemy.Table
        Proxy object reflecting table with found discrapancies from B.
    fetched : int or None
        Number of fetched records from data source.
    success : int or None
        Number of success results.
    errors : int or None
        Number of discrapancies.
    error_level : float or None
        Indicator presenting the percentage of discrapancies among the fetched
        records.
    fetched_a : int or None
        Number of fetched records from data source A.
    fetched_b : int or None
        Number of fetched records from data source B.
    success_a : int or None
        Number of success results in A.
    success_b : int or None
        Number of success results in B.
    errors_a : int or None
        Number of discrapancies in A.
    errors_b : int or None
        Number of discrapancies in B.
    error_level_a : float or None
        Indicator presenting the percentage of discrapancies among the fetched
        records from A.
    error_level_b : float or None
        Indicator presenting the percentage of discrapancies among the fetched
        records from B.
    date_from : datetime
        Data source date lower bound.
    date_from : datetime
        Data source date upper bound.
    """

    def __init__(self, name=None, timestamp=None, process_id=None,
                 date=None, date_from=None, date_to=None):
        self.name = name or reader.read_control_name(process_id)
        self.config = reader.read_control_config(self.name)
        self.result = reader.read_control_result(process_id)
        self.id = int(self.config['control_id'])
        self.group = self.config['control_group']
        self.type = self.config['control_type']
        self.subtype = self.config['control_subtype']
        self.engine = self.config['control_engine']
        self.timestamp = timestamp

        self.parser = Parser(self)
        self.executor = Executor(self)

        self.process_id = process_id
        if self.result:
            self.start_date = self.result['start_date']
            self.end_date = self.result['end_date']
            self.updated = self.result['updated']
            self.status = self.result['status']

            self.date_from = self.result['date_from']
            self.date_to = self.result['date_to']

            self.fetched = self.result['fetched']
            self.success = self.result['success']
            self.errors = self.result['errors']
            self.error_level = self.result['error_level']

            self.fetched_a = self.result['fetched_a']
            self.fetched_b = self.result['fetched_b']
            self.success_a = self.result['success_a']
            self.success_b = self.result['success_b']
            self.errors_a = self.result['errors_a']
            self.errors_b = self.result['errors_b']
            self.error_level_a = self.result['error_level_a']
            self.error_level_b = self.result['error_level_b']
        else:
            self.start_date = None
            self.end_date = None
            self.updated = None
            self.status = None

            if self.timestamp:
                self.date_from, self.date_to = self.parser.parse_dates()
            elif date:
                self.date_from = self.parser.parse_date(date, 0, 0, 0)
                self.date_to = self.parser.parse_date(date, 23, 59, 59)
            else:
                self.date_from = self.parser.parse_date(date_from)
                self.date_to = self.parser.parse_date(date_to)

            self.fetched = None
            self.success = None
            self.errors = None
            self.error_level = None

            self.fetched_a = None
            self.success_a = None
            self.errors_a = None
            self.error_level_a = None

            self.fetched_b = None
            self.success_b = None
            self.errors_b = None
            self.error_level_b = None

        self.source_table = None
        self.source_table_a = None
        self.source_table_b = None

        self.input_table = None
        self.input_table_a = None
        self.input_table_b = None

        self.result_table = None
        self.error_table = None
        self.error_table_a = None
        self.error_table_b = None

        self.output_table = None
        self.output_table_a = None
        self.output_table_b = None

        self._with_error = False
        self._all_errors = []
        self._pending_error = None
        pass

    def __str__(self):
        """Take control information and represent it as a simple string.

        Returns
        -------
        value : str
            Control name with or withoud process id.
        """
        if self.process_id is None:
            return f'[{self.name}]'
        else:
            return f'[{self.name}:{self.process_id}]'
        pass

    __repr__ = __str__

    @property
    def name(self):
        """Get control name."""
        return self._name

    @name.setter
    def name(self, value):
        if isinstance(value, str) or value is None:
            self._name = value
        else:
            type = value.__class__.__name__
            message = f'name must be str or None, not {type}'
            raise TypeError(message)
        pass

    @property
    def process_id(self):
        """Get control run process ID."""
        return self._process_id

    @property
    def pid(self):
        """Shortcut for process_id."""
        return self._process_id

    @process_id.setter
    def process_id(self, value):
        if isinstance(value, int) or value is None:
            self._process_id = value
        else:
            type = value.__class__.__name__
            message = f'process_id must be int or None, not {type}'
            raise TypeError(message)
        pass

    @property
    def source_name(self):
        """Get control data source name."""
        return utils.to_lower(self.config['source_name'])

    @property
    def source_date_field(self):
        """Get control data source date field."""
        return utils.to_lower(self.config['source_date_field'])

    @property
    def source_name_a(self):
        """Get control data source A name."""
        return utils.to_lower(self.config['source_name_a'])

    @property
    def source_date_field_a(self):
        """Get control data source A date field."""
        return utils.to_lower(self.config['source_date_field_a'])

    @property
    def source_name_b(self):
        """Get control data source B name."""
        return utils.to_lower(self.config['source_name_b'])

    @property
    def source_date_field_b(self):
        """Get control data source B date field."""
        return utils.to_lower(self.config['source_date_field_b'])

    @property
    def status(self):
        """Get control run status."""
        return self._status

    @status.setter
    def status(self, value):
        """Set control run status."""
        if isinstance(value, str) and len(value) == 1:
            if hasattr(self, '_status'):
                self._status = value
                self.updated = dt.datetime.now()
                logger.info(f'{self} Status changed to {self._status}')
            else:
                self._status = value
        elif value is None:
            self._status = None
        else:
            message = f'incorrect status: {value}'
            raise ValueError(message)
        pass

    @property
    def initiated(self):
        """Check if control is initiated."""
        if self.status == 'I':
            return True
        else:
            return False
        pass

    @property
    def working(self):
        """Check if control is working."""
        if self.status in ('S', 'P', 'F'):
            return True
        else:
            return False
        pass

    @property
    def select(self):
        """Get SQL statement to fetch data from data source."""
        return self.parser.parse_select()

    @property
    def select_a(self):
        """Get SQL statement to fetch data from data source A."""
        return self.parser.parse_select_a()

    @property
    def select_b(self):
        """Get SQL statement to fetch data from data source B."""
        return self.parser.parse_select_b()

    @property
    def output_names(self):
        """Get output table names."""
        return self.parser.parse_output_names()

    @property
    def output_tables(self):
        """Get output tables."""
        return self.parser.parse_output_tables()

    @property
    def temp_names(self):
        """Get temporary table names."""
        return self.parser.parse_temp_names()

    @property
    def temp_tables(self):
        """Get temporary tables."""
        return self.parser.parse_temp_tables()

    @property
    def rule_config(self):
        """Get control match configuration."""
        if self.type == 'ANL':
            return []
        elif self.type == 'REC':
            return self.parser.parse_reconciliation_rule_config()
        elif self.type == 'REP':
            return []
        pass

    @property
    def error_config(self):
        """Get control error configuration."""
        if self.type == 'ANL':
            return self.parser.parse_analyze_error_config()
        elif self.type == 'REC':
            return self.parser.parse_reconciliation_error_config()
        elif self.type == 'REP':
            return []
        pass

    @property
    def output_columns(self):
        """Get control output column configuration."""
        return self.parser.parse_output_columns()

    @property
    def output_columns_a(self):
        """Get control output A column configuration."""
        return self.parser.parse_output_columns_a()

    @property
    def output_columns_b(self):
        """Get control output B column configuration."""
        return self.parser.parse_output_columns_b()

    @property
    def need_a(self):
        """Get parameter defining necessity of data source A saving."""
        return True if self.config['need_a'] == 'Y' else False

    @property
    def need_b(self):
        """Get parameter defining necessity of data source B saving."""
        return True if self.config['need_b'] == 'Y' else False

    @property
    def need_hook(self):
        """Get parameter defining necessity of hook execution."""
        return True if self.config['need_hook'] == 'Y' else False

    @property
    def text_error(self):
        """Get textual error representation of current control run."""
        errors = self._all_errors.copy()
        texts = [''.join(tb.format_exception(*error)) for error in errors]
        text = f'{str():->40}\n'.join(texts)
        return text

    def run(self):
        """Run control in an ordinary way."""
        logger.debug(f'{self} Running control...')
        if self._initiate() is True:
            if self._start() is True:
                if self._progress() is True:
                    if self._finish() is True:
                        if self._done() is True:
                            self._hook()
        pass

    def process(self):
        """Run control as a separate accompanied stoppable process."""
        logger.debug(f'{self} Running control...')
        if self._initiate() is True:
            self._resume()
        pass

    def resume(self):
        """Resume initiated control run."""
        if self.initiated:
            if self._start() is True:
                if self._progress() is True:
                    if self._finish() is True:
                        if self._done() is True:
                            self._hook()
        pass

    def cancel(self):
        """Cancel control run."""
        if self.working:
            self._terminate()
        pass

    def delete(self):
        """Delete control results."""
        self._delete()
        pass

    def revoke(self):
        """Revoke control results."""
        self._revoke()
        pass

    def clean(self):
        """Clean control results."""
        self._clean()
        pass

    def _initiate(self):
        logger.info(f'{self} Initiating control...')
        try:
            self.status = 'I'
            logger.debug(f'{self} Creating new record in {db.tables.log}')
            conn = db.connect()
            insert = db.tables.log.insert()
            insert = insert.values(control_id=self.id,
                                   added=dt.datetime.now(),
                                   status=self.status,
                                   date_from=self.date_from,
                                   date_to=self.date_to)
            result = conn.execute(insert)
            self.process_id = int(result.inserted_primary_key[0])
        except Exception:
            logger.error()
            return self._exit()
        else:
            logger.debug(f'{self} New record in {db.tables.log} created')
            logger.info(f'{self} Control owns process ID {self.process_id}')
            logger.info(f'{self} Control initiated')
            return self._continue()
        pass

    def _resume(self):
        process = mp.Process(target=self.resume)
        process.start()
        logger.info(f'{self} Running as process on PID {process.pid}')
        while process.is_alive():
            tm.sleep(5)
            control = Control(process_id=self.process_id)
            if control.status is None:
                logger.info(f'{self} Process termination request received')
                process.terminate()
                logger.info(f'{self} Process PID {process.pid} terminated')
                self._cancel()
                continue
        process.join()
        self.__dict__.update(control.__dict__)
        pass

    def _start(self):
        logger.info(f'{self} Starting control...')
        try:
            self.status = 'S'
            self.start_date = dt.datetime.now()
            self._update(status=self.status, start_date=self.start_date)
            if self.type in ('ANL', 'REP'):
                self.source_table = self.parser.parse_source_table()
            elif self.type == 'REC':
                self.source_table_a = self.parser.parse_source_table_a()
                self.source_table_b = self.parser.parse_source_table_b()
        except Exception:
            logger.error()
            return self._exit()
        else:
            logger.info(f'{self} Control started at {self.start_date}')
            return self._continue()
        pass

    def _progress(self):
        try:
            self.status = 'P'
            self._update(status=self.status)
            self._fetch()
            self._execute()
            self._save()
        except Exception:
            logger.error()
            return self._exit()
        else:
            return self._continue()
        pass

    def _finish(self):
        logger.info(f'{self} Finishing control...')
        try:
            self.status = 'F'
            self._update(status=self.status)
            self.executor.drop_temporary_tables()
        except Exception:
            logger.error()
            return self._exit()
        else:
            logger.info(f'{self} Control finished')
            return self._continue()
        pass

    def _cancel(self):
        logger.info(f'{self} Canceling control...')
        try:
            self.status = 'C'
            self._update(status=self.status)
            self.executor.drop_temporary_tables()
            self.executor.delete_output_records()
        except Exception:
            logger.error()
        else:
            logger.info(f'{self} Control canceled')
        pass

    def _done(self):
        try:
            self.status = 'D'
            self.end_date = dt.datetime.now()
            self._update(status=self.status, end_date=self.end_date)
        except Exception:
            logger.error()
            return self._exit()
        else:
            logger.info(f'{self} ended at {self.end_date}')
            return self._continue()
        pass

    def _error(self):
        try:
            self.status = 'E'
            self.end_date = dt.datetime.now()
            self._update(status=self.status, end_date=self.end_date)
        except Exception:
            logger.error()
        else:
            logger.info(f'{self} ended with error at {self.end_date}')
        pass

    def _continue(self):
        return True

    def _exit(self):
        self._with_error = True
        self._all_errors.append(sys.exc_info())
        self._update(text_error=self.text_error)
        return self._error()

    def _fetch(self):
        logger.info(f'{self} Fetching records...')
        if self.type in ('ANL', 'REP'):
            logger.info(f'{self} Fetching {self.source_name}...')
            self.input_table = self.executor.fetch_records()
            self.fetched = self.executor.count_fetched()
            logger.info(f'{self} Records fetched: {self.fetched}')
            self._update(fetched=self.fetched)
        elif self.type == 'REC':
            threads = []
            current = th.current_thread()
            for attr in ('_fetch_a', '_fetch_b'):
                name = f'{current.name}({attr[1:]})'
                func = getattr(self, attr)
                thread = th.Thread(target=func, name=name, daemon=True)
                thread.start()
                threads.append(thread)
            for thread in threads:
                thread.join()
                if self._pending_error is not None:
                    raise self._pending_error
            self._update(fetched_a=self.fetched_a, fetched_b=self.fetched_b)
        pass

    def _fetch_a(self):
        try:
            self.__fetch_a()
        except Exception as error:
            self._pending_error = error
        pass

    def __fetch_a(self):
        logger.info(f'{self} Fetching {self.source_name_a}...')
        self.input_table_a = self.executor.fetch_records_a()
        self.fetched_a = self.executor.count_fetched_a()
        logger.info(f'{self} Records fetched A: {self.fetched_a}')
        pass

    def _fetch_b(self):
        try:
            self.__fetch_b()
        except Exception as error:
            self._pending_error = error
        pass

    def __fetch_b(self):
        logger.info(f'{self} Fetching {self.source_name_b}...')
        self.input_table_b = self.executor.fetch_records_b()
        self.fetched_b = self.executor.count_fetched_b()
        logger.info(f'{self} Records fetched B: {self.fetched_b}')
        pass

    def _execute(self):
        logger.info(f'{self} Executing control...')
        if self.type == 'ANL':
            if self.fetched or 0 > 0:
                self.error_table = self.executor.analyze()
                self.errors = self.executor.count_errors()
                self.success = self.fetched-self.errors
                self.error_level = (self.errors/self.fetched)*100
                self._update(errors=self.errors,
                             success=self.success,
                             error_level=self.error_level)
        elif self.type == 'REP':
            if self.fetched or 0 > 0:
                self.error_table = self.executor.analyze()
        elif self.type == 'REC' and self.subtype == 'MA':
            threads = []
            current = th.current_thread()
            for action in ('match', 'mismatch'):
                name = f'{current.name}({action})'
                func = getattr(self, f'_{action}')
                thread = th.Thread(target=func, name=name, daemon=True)
                thread.start()
                threads.append(thread)
            for thread in threads:
                thread.join()
                if self._pending_error is not None:
                    raise self._pending_error
            self.error_level = (self.errors/(self.success+self.errors))*100
            self._update(errors=self.errors,
                         success=self.success,
                         error_level=self.error_level)
        logger.info(f'{self} Control executed')
        pass

    def _match(self):
        try:
            self.__match()
        except Exception as error:
            self._pending_error = error
        pass

    def __match(self):
        self.result_table = self.executor.match()
        self.success = self.executor.count_matched()
        pass

    def _mismatch(self):
        try:
            self.__mismatch()
        except Exception as error:
            self._pending_error = error
        pass

    def __mismatch(self):
        self.error_table = self.executor.mismatch()
        self.errors = self.executor.count_mismatched()
        pass

    def _save(self):
        logger.info(f'{self} Saving results...')
        if self.type == 'ANL':
            if self.errors or 0 > 0:
                self.executor.save_errors()
        elif self.type == 'REP':
            if self.fetched or 0 > 0:
                self.executor.save_errors()
        elif self.type == 'REC':
            if self.subtype == 'MA':
                if self.errors or 0 > 0:
                    self.executor.save_mismatches()
        logger.info(f'{self} Results saved')
        pass

    def _terminate(self):
        logger.info(f'{self} Creating process termination request...')
        try:
            self.status = None
            self._update(status=self.status)
        except Exception:
            logger.error()
        else:
            logger.info(f'{self} Process termination request created')
        pass

    def _delete(self):
        logger.info(f'{self} Deleting results...')
        self.executor.delete_output_records()
        logger.info(f'{self} Results deleted')
        pass

    def _revoke(self):
        try:
            logger.info(f'{self} Revoking control...')
            self.status = 'X'
            self._update(status=self.status)
            self._delete()
        except Exception:
            logger.error()
        else:
            logger.info(f'{self} Control revoked')
        pass

    def _clean(self):
        logger.info(f'{self} Cleaning control results...')
        conn = db.connect()
        outdated_results = list(self.parser.parse_outdated_results())
        if outdated_results:
            for table, process_ids in outdated_results:
                for process_id in process_ids:
                    repr = f'[{self.name}:{process_id}]'
                    logger.info(f'{repr} Deleting results in {table}...')
                    id = table.c.rapo_process_id
                    query = table.delete().where(id == process_id)
                    text = db.formatter(query)
                    logger.debug(f'{self} Deleting from {table} '
                                 f'with query:\n{text}')
                    conn.execute(query)
                    logger.info(f'{repr} Results in {table} deleted')
            logger.info(f'{self} Control results cleaned')
        else:
            logger.info(f'{self} No control results to clean')
        pass

    def _update(self, **kwargs):
        logger.debug(f'{self} Updating {db.tables.log} with {kwargs}')
        conn = db.connect()
        update = db.tables.log.update()
        update = update.values(**kwargs, updated=dt.datetime.now())
        update = update.where(db.tables.log.c.process_id == self.process_id)
        conn.execute(update)
        logger.debug(f'{self} {db.tables.log} updated')
        pass

    def _hook(self):
        if self.need_hook is True:
            self.executor.hook()
        pass

    pass


class Parser():
    """Represents control parser."""

    def __init__(self, owner):
        self.__owner = owner
        pass

    @property
    def control(self):
        """Get owning control instance."""
        return self.__owner

    @property
    def c(self):
        """Shortcut for control."""
        return self.__owner

    def parse_dates(self):
        """."""
        days_back = self.control.config['days_back']
        date_from = self.parse_date_from()-dt.timedelta(days=days_back)
        date_to = self.parse_date_to()-dt.timedelta(days=days_back)
        return (date_from, date_to)

    def parse_date_from(self):
        """Get data source date lower bound.

        Returns
        -------
        date_from : datetime or None
            Fetched records from data source should begin from this date.
        """
        timestamp = self.control.timestamp
        date = self._parse_date(timestamp, 0, 0, 0)
        return date

    def parse_date_to(self):
        """Get data source date upper bound.

        date_to : datetime or None
            Fetched records from data source should end with this date.
        """
        timestamp = self.control.timestamp
        date = self._parse_date(timestamp, 23, 59, 59)
        return date

    def parse_date(self, value, hour=None, minute=None, second=None):
        """Get date from initial raw value."""
        return self._parse_date(value, hour=hour, minute=minute, second=second)

    def _parse_date(self, value, hour=None, minute=None, second=None):
        date = utils.to_date(value)
        if hour is not None or minute is not None or second is not None:
            kwargs = {'hour': hour, 'minute': minute, 'second': second}
            kwargs = {k: v for k, v in kwargs.items() if v is not None}
            date = date.replace(**kwargs)
        return date

    def parse_source_table(self):
        """Get data source table.

        Returns
        -------
        table : sqlalchemy.Table
            Object reflecting data source table.
        """
        name = self.control.source_name
        table = self._parse_source_table(name)
        return table

    def parse_source_table_a(self):
        """Get data source A table.

        Returns
        -------
        table : sqlalchemy.Table
            Object reflecting data source A table.
        """
        name = self.control.source_name_a
        table = self._parse_source_table(name)
        return table

    def parse_source_table_b(self):
        """Get data source B table.

        Returns
        -------
        table : sqlalchemy.Table
            Object reflecting data source B table.
        """
        name = self.control.source_name_b
        table = self._parse_source_table(name)
        return table

    def _parse_source_table(self, name):
        logger.debug(f'{self.c} Parsing source table {name}...')
        if isinstance(name, str) is True or name is None:
            table = db.table(name) if isinstance(name, str) is True else None
            if table is None:
                message = f'Source table {name} is not defined'
                raise AttributeError(message)
            else:
                logger.debug(f'{self.c} Source table {name} parsed')
                return table
        else:
            type = name.__class__.__name__
            message = f'source name must be str or None not {type}'
            raise TypeError(message)

    def parse_select(self):
        """Get SQL statement to fetch data from data source.

        Returns
        -------
        select : sqlalchemy.Select
        """
        table = self.control.source_table
        date_field = self.control.source_date_field
        select = self._parse_select(table, date_field=date_field)
        return select

    def parse_select_a(self):
        """Get SQL statement to fetch data from data source A.

        Returns
        -------
        select : sqlalchemy.Select
        """
        table = self.control.source_table_a
        date_field = self.control.source_date_field_a
        select = self._parse_select(table, date_field=date_field)
        return select

    def parse_select_b(self):
        """Get SQL statement to fetch data from data source B.

        Returns
        -------
        select : sqlalchemy.Select
        """
        table = self.control.source_table_b
        date_field = self.control.source_date_field_b
        select = self._parse_select(table, date_field=date_field)
        return select

    def _parse_select(self, table, date_field=None):
        logger.debug(f'{self.c} Parsing {table} select...')
        select = table.select()
        if isinstance(date_field, str) is True:
            column = table.c[date_field]

            datefmt = '%Y-%m-%d %H:%M:%S'
            date_from = self.control.date_from.strftime(datefmt)
            date_to = self.control.date_to.strftime(datefmt)

            datefmt = 'YYYY-MM-DD HH24:MI:SS'
            date_from = sa.func.to_date(date_from, datefmt)
            date_to = sa.func.to_date(date_to, datefmt)

            select = select.where(column.between(date_from, date_to))
            logger.debug(f'{self.c} {table} select parsed')
        return select

    def parse_output_names(self):
        """Get list with necessary output table names.

        Returns
        -------
        names : list
            List necessary output names.
        """
        names = []
        if (
            self.control.type == 'ANL'
            or (
                self.control.type == 'REC'
                and self.control.subtype == 'MA'
            )
            or self.control.type == 'REP'
        ):
            name = f'rapo_rest_{self.control.name}'.lower()
            names.append(name)
        return names

    def parse_output_tables(self):
        """Get list with existing output tables.

        Returns
        -------
        tables : list
            List of sqlalchemy.Table objects.
        """
        tables = []
        for name in self.control.output_names:
            if db.engine.has_table(name):
                table = db.table(name)
                tables.append(table)
        return tables

    def parse_temp_names(self):
        """Get list with necessary temporary table names.

        Returns
        -------
        names : list
            List necessary temporary names.
        """
        names = []
        fd = f'rapo_temp_fd_{self.control.process_id}'
        fda = f'rapo_temp_fda_{self.control.process_id}'
        fdb = f'rapo_temp_fdb_{self.control.process_id}'
        err = f'rapo_temp_err_{self.control.process_id}'
        md = f'rapo_temp_md_{self.control.process_id}'
        nmd = f'rapo_temp_nmd_{self.control.process_id}'
        if self.control.type == 'ANL':
            names.extend([fd, err])
        elif self.control.type == 'REP':
            names.append(fd)
        elif self.control.type == 'REC' and self.control.subtype == 'MA':
            names.extend([fda, fdb, md, nmd])
        return names

    def parse_temp_tables(self):
        """Get list with existing temporary tables.

        Returns
        -------
        tables : list
            List of sqlalchemy.Table objects.
        """
        tables = []
        for name in self.control.temp_names:
            if db.engine.has_table(name):
                table = db.table(name)
                tables.append(table)
        return tables

    def parse_analyze_error_config(self):
        """Get analyze error configuration.

        Returns
        -------
        config : list
            List with dictionaries where presented all keys for discrapancies.
        """
        logger.debug(f'{self.c} Parsing error configuration...')
        raw = self.control.config['error_config']
        config = []
        for item in json.loads(raw or '[]'):
            connexion = item.get('connexion', 'and').upper()
            column = item.get('column', '').lower()
            column_a = item.get('column_a', '').lower()
            column_b = item.get('column_b', '').lower()
            relation = item.get('relation', '<>').upper()
            value = item.get('value')
            is_column = item.get('is_column', False)

            config.append(
                {'connexion': connexion,
                 'column': column or None,
                 'column_a': column_a or None,
                 'column_b': column_b or None,
                 'relation': relation,
                 'value': value.lower() if is_column is True else value,
                 'is_column': is_column})
        logger.debug(f'{self.c} Error configuration parsed')
        return config

    def parse_reconciliation_rule_config(self):
        """Get control rule configuration.

        Returns
        -------
        config : list
            List with dictionaries where presented all keys for rule.
        """
        logger.debug(f'{self.c} Parsing rule configuration...')
        raw = self.control.config['rule_config']
        config = []
        for item in json.loads(raw or '[]'):
            column_a = item['column_a'].lower()
            column_b = item['column_b'].lower()

            new = {'column_a': column_a, 'column_b': column_b}
            config.append(new)
        logger.debug(f'{self.c} Rule configuration parsed')
        return config

    def parse_reconciliation_error_config(self):
        """Get reconciliation error configuration.

        Returns
        -------
        config : list
            List with dictionaries where presented all keys for mismatching.
        """
        logger.debug(f'{self.c} Parsing error configuration...')
        raw = self.control.config['error_config']
        config = []
        for item in json.loads(raw or '[]'):
            column_a = item['column_a'].lower()
            column_b = item['column_b'].lower()

            new = {'column_a': column_a, 'column_b': column_b}
            config.append(new)
        logger.debug(f'{self.c} Error configuration parsed')
        return config

    def parse_output_columns(self):
        """Get control output columns.

        Returns
        -------
        columns : list or None
            List with dictionaries where output columns configuration is
            presented.
            Will be None if configuration is not filled in RAPO_CONFIG.
        """
        config = self.control.config['output_table']
        columns = self._parse_output_columns(config)
        return columns

    def parse_output_columns_a(self):
        """Get control output columns A.

        Returns
        -------
        columns : list or None
            List with dictionaries where output A columns configuration is
            presented.
            Will be None if configuration is not filled in RAPO_CONFIG.
        """
        config = self.control.config['output_table_a']
        columns = self._parse_output_columns(config)
        return columns

    def parse_output_columns_b(self):
        """Get control output columns B.

        Returns
        -------
        columns : list or None
            List with dictionaries where output B columns configuration is
            presented.
            Will be None if configuration is not filled in RAPO_CONFIG.
        """
        config = self.control.config['output_table_b']
        columns = self._parse_output_columns(config)
        return columns

    def _parse_output_columns(self, config):
        logger.debug(f'{self.c} Parsing output columns...')
        config = json.loads(config) if isinstance(config, str) else None
        if config is None:
            logger.debug(f'{self.c} Output was not configured')
            return None
        else:
            column = dict.fromkeys(['column', 'column_a', 'column_b'])
            columns = []
            for value in config.get('columns', []):
                new = column.copy()
                if isinstance(value, str) is True:
                    new['column'] = value.lower()
                if isinstance(value, dict) is True:
                    for key in new.keys():
                        raw = value.get(key)
                        if isinstance(raw, str):
                            new[key] = raw.lower()
                columns.append(new)
            if columns:
                logger.debug(f'{self.c} Output columns parsed')
                return columns
            else:
                logger.debug(f'{self.c} Output columns was not configured')
                return None
        pass

    def parse_outdated_results(self):
        """Create list with tables and IDs of outdated control results."""
        conn = db.connect()
        log = db.tables.log
        control_id = self.control.id
        days_retention = self.control.config['days_retention']
        for table in self.parse_output_tables():
            select = sa.select([log.c.process_id])
            date = sa.func.trunc(sa.func.sysdate())-days_retention
            subq = sa.select([table.c.rapo_process_id])
            query = (select.where(log.c.control_id == control_id)
                           .where(log.c.added < date)
                           .where(log.c.process_id.in_(subq))
                           .order_by(log.c.process_id))
            text = db.formatter(query)
            logger.debug(f'{self.c} Searching outdated results in {table} '
                         f'with query:\n{text}')
            result = conn.execute(query)
            pids = [row[0] for row in result]
            logger.debug(f'{self.c} Outdated results in {table}: {pids}')
            if pids:
                yield (table, pids)
        pass

    pass


class Executor():
    """Represents control executor."""

    def __init__(self, bind):
        self.__bind = bind
        pass

    @property
    def control(self):
        """Get binded control instance."""
        return self.__bind

    @property
    def c(self):
        """Get binded control instance. Same as control property."""
        return self.__bind

    def fetch_records(self):
        """Fetch data source.

        Returns
        -------
        table : sqlalchemy.Table
            Object reflecting table with fetched data in case if DB is chosen
            engine.
        """
        select = self.control.select
        if self.control.engine == 'DB':
            tablename = f'rapo_temp_fd_{self.control.process_id}'
            table = self._fetch_records_to_table(select, tablename)
            return table

    def fetch_records_a(self):
        """Fetch data source A.

        Returns
        -------
        table : sqlalchemy.Table
            Object reflecting table with fetched data in case if DB is chosen
            engine.
        """
        select = self.control.select_a
        if self.control.engine == 'DB':
            tablename = f'rapo_temp_fda_{self.control.process_id}'
            table = self._fetch_records_to_table(select, tablename)
            return table

    def fetch_records_b(self):
        """Fetch data source B.

        Returns
        -------
        table : sqlalchemy.Table
            Object reflecting table with fetched data in case if DB is chosen
            engine.
        """
        select = self.control.select_b
        if self.control.engine == 'DB':
            tablename = f'rapo_temp_fdb_{self.control.process_id}'
            table = self._fetch_records_to_table(select, tablename)
            return table

    def analyze(self):
        """Run data analyze used for control with ANL type.

        Returns
        -------
        table : sqlalchemy.Table
            Object reflecting table with found discrepancies in case if DB is
            chosen engine.
        """
        logger.debug(f'{self.c} Analyzing...')
        conn = db.connect()
        input_table = self.control.input_table
        output_columns = self.control.output_columns
        if output_columns is None or len(output_columns) == 0:
            select = input_table.select()
        else:
            columns = []
            for output_column in output_columns:
                name = output_column['column']
                column = input_table.c[name]
                columns.append(column)
            select = sa.select(columns)
        texts = []
        config = self.control.error_config
        for item in config:
            text = []
            connexion = item['connexion']
            if len(texts) > 0:
                text.append(connexion)
            column = str(input_table.c[item['column']])
            text.append(column)
            relation = item['relation']
            text.append(relation)
            value = item['value']
            is_column = item['is_column']
            value = str(input_table.c[value]) if is_column is True else value
            text.append(value)
            text = ' '.join(text)
            texts.append(text)
        select = select.where(sa.text('\n'.join(texts)))

        tablename = f'rapo_temp_err_{self.control.process_id}'
        select = select.compile(bind=db.engine, compile_kwargs=db.ckwargs)
        ctas = sa.text(f'CREATE TABLE {tablename} AS\n{select}')
        text = db.formatter(ctas)
        logger.info(f'{self.c} Creating {tablename} with query:\n{text}')
        conn.execute(ctas)
        logger.debug(f'{self.c} {tablename} created')

        table = db.table(tablename)
        logger.debug(f'{self.c} Analyzing done')
        return table

    def match(self):
        """Run data matching for control with REC type and MA subtype.

        Returns
        -------
        table : sqlalchemy.Table
            Object reflecting table with matched data in case if DB is chosen
            engine.
        """
        logger.debug(f'{self.c} Defining matches...')

        conn = db.connect()
        table_a = self.control.input_table_a
        table_b = self.control.input_table_b

        columns = []
        output_columns = self.control.output_columns
        if output_columns is None or len(output_columns) == 0:
            columns.extend(table_a.columns)
            columns.extend(table_b.columns)
        else:
            for output_column in output_columns:
                name = output_column['column']

                column_a = output_column['column_a']
                column_a = table_a.c[column_a] if column_a else None

                column_b = output_column['column_b']
                column_b = table_b.c[column_b] if column_b else None

                if column_a is not None and column_b is not None:
                    column = sa.func.coalesce(column_a, column_b)
                elif column_a is not None:
                    column = column_a
                elif column_b is not None:
                    column = column_b

                column = column.label(name) if name else column
                columns.append(column)

        keys = []
        for rule in self.control.rule_config:
            column_a = table_a.c[rule['column_a']]
            column_b = table_b.c[rule['column_b']]
            keys.append(column_a == column_b)
        join = table_a.join(table_b, *keys)
        select = sa.select(columns).select_from(join)

        keys = []
        for error in self.control.error_config:
            column_a = table_a.c[error['column_a']]
            column_b = table_b.c[error['column_b']]
            select = select.where(column_a == column_b)

        tablename = f'rapo_temp_md_{self.control.process_id}'
        select = select.compile(bind=db.engine, compile_kwargs=db.ckwargs)
        ctas = sa.text(f'CREATE TABLE {tablename} AS\n{select}')
        text = db.formatter(ctas)
        logger.info(f'{self.c} Creating {tablename} with query:\n{text}')
        conn.execute(ctas)
        logger.debug(f'{self.c} {tablename} created')

        table = db.table(tablename)
        logger.debug(f'{self.c} Matches defined')
        return table

    def mismatch(self):
        """Run data mismatching for control with REC type and MA subtype.

        Returns
        -------
        table : sqlalchemy.Table
            Object reflecting table with mismatched data in case if DB is
            chosen engine.
        """
        logger.debug(f'{self.c} Defining mismatches...')

        conn = db.connect()
        table_a = self.control.input_table_a
        table_b = self.control.input_table_b

        columns = []
        output_columns = self.control.output_columns
        if output_columns is None or len(output_columns) == 0:
            columns.extend(table_a.columns)
            columns.extend(table_b.columns)
        else:
            for output_column in output_columns:
                name = output_column['column']

                column_a = output_column['column_a']
                column_a = table_a.c[column_a] if column_a else None

                column_b = output_column['column_b']
                column_b = table_b.c[column_b] if column_b else None

                if column_a is not None and column_b is not None:
                    column = sa.func.coalesce(column_a, column_b)
                elif column_a is not None:
                    column = column_a
                elif column_b is not None:
                    column = column_b

                column = column.label(name) if name else column
                columns.append(column)

        keys = []
        for rule in self.control.rule_config:
            column_a = table_a.c[rule['column_a']]
            column_b = table_b.c[rule['column_b']]
            keys.append(column_a == column_b)
        join = table_a.join(table_b, *keys)
        select = sa.select(columns).select_from(join)

        keys = []
        for error in self.control.error_config:
            column_a = table_a.c[error['column_a']]
            column_b = table_b.c[error['column_b']]
            select = select.where(column_a != column_b)

        tablename = f'rapo_temp_nmd_{self.control.process_id}'
        select = select.compile(bind=db.engine, compile_kwargs=db.ckwargs)
        ctas = sa.text(f'CREATE TABLE {tablename} AS\n{select}')
        text = db.formatter(ctas)
        logger.info(f'{self.c} Creating {tablename} with query:\n{text}')
        conn.execute(ctas)
        logger.debug(f'{self.c} {tablename} created')

        table = db.table(tablename)
        logger.debug(f'{self.c} Mismatches defined')
        return table

    def count_fetched(self):
        """Count fetched records in data source."""
        if self.control.engine == 'DB':
            table = self.control.input_table
            fetched = self._count_fetched_to_table(table)
        return fetched

    def count_fetched_a(self):
        """Count fetched records in data source A."""
        if self.control.engine == 'DB':
            table = self.control.input_table_a
            fetched_a = self._count_fetched_to_table(table)
        return fetched_a

    def count_fetched_b(self):
        """Count fetched records in data source B."""
        if self.control.engine == 'DB':
            table = self.control.input_table_b
            fetched_b = self._count_fetched_to_table(table)
        return fetched_b

    def count_errors(self):
        """Count found discrepancies for control with ANL type.

        Returns
        -------
        errors : int
            Number of found discrepancies.
        """
        logger.debug(f'{self.c} Counting errors...')
        if self.control.engine == 'DB':
            conn = db.connect()
            table = self.control.error_table
            count = sa.select([sa.func.count()]).select_from(table)
            errors = conn.execute(count).scalar()
        logger.debug(f'{self.c} Errors counted')
        return errors

    def count_matched(self):
        """Count records that were matched.

        Returns
        -------
        matched : int
            Number of found discrepancies.
        """
        logger.debug(f'{self.c} Counting matched...')
        if self.control.engine == 'DB':
            conn = db.connect()
            table = self.control.result_table
            count = sa.select([sa.func.count()]).select_from(table)
            matched = conn.execute(count).scalar()
        logger.debug(f'{self.c} Matched counted')
        return matched

    def count_mismatched(self):
        """Count records that were not matched.

        Returns
        -------
        mismatched : int
            Number of found discrepancies.
        """
        logger.debug(f'{self.c} Counting mismatched')
        if self.control.engine == 'DB':
            conn = db.connect()
            table = self.control.error_table
            count = sa.select([sa.func.count()]).select_from(table)
            mismatched = conn.execute(count).scalar()
        logger.debug(f'{self.c} Mismatched counted')
        return mismatched

    def save_results(self):
        """Save defined results as output records."""
        logger.debug(f'{self.c} Start saving...')
        conn = db.connect()
        table = self.control.result_table
        process_id = sa.literal(self.control.process_id)
        select = sa.select([*table.columns,
                            process_id.label('rapo_process_id')])
        table = self.prepare_output_table()
        insert = table.insert().from_select(table.columns, select)
        conn.execute(insert)
        logger.debug(f'{self.c} Saving done')
        pass

    def save_errors(self):
        """Save defined errors as output records."""
        logger.debug(f'{self.c} Start saving...')
        conn = db.connect()
        table = self.control.error_table
        process_id = sa.literal(self.control.process_id)
        select = sa.select([*table.columns,
                            process_id.label('rapo_process_id')])
        table = self.prepare_output_table()
        insert = table.insert().from_select(table.columns, select)
        conn.execute(insert)
        logger.debug(f'{self.c} Saving done')
        pass

    def save_matches(self):
        """Save found matches as RAPO results."""
        return self.save_results()

    def save_mismatches(self):
        """Save found mismatches as RAPO results."""
        return self.save_errors()

    def delete_output_records(self):
        """Delete records saved as control results in DB table."""
        conn = db.connect()
        for table in self.control.output_tables:
            id = table.c.rapo_process_id
            delete = table.delete().where(id == self.control.process_id)
            conn.execute(delete)
        pass

    def prepare_output_table(self):
        """Check RAPO_RESULT and create it at initial control run.

        Returns
        -------
        table : sqlalchemy.Table
            Object reflecting RAPO_RESULT.
        """
        tablename = f'rapo_rest_{self.control.name}'.lower()
        conn = db.connect()
        if conn.engine.has_table(tablename) is False:
            logger.debug(f'{self.c} Table {tablename} will be created')

            columns = []
            output_columns = self.control.output_columns
            if output_columns is None or len(output_columns) == 0:
                if self.c.config['source_name'] is not None:
                    columns.extend(self.c.source_table.columns)
                if self.c.config['source_name_a'] is not None:
                    columns.extend(self.c.source_table_a.columns)
                if self.c.config['source_name_b'] is not None:
                    columns.extend(self.c.source_table_b.columns)
            else:
                for output_column in output_columns:
                    name = output_column['column']
                    column_a = output_column['column_a']
                    column_b = output_column['column_b']
                    if column_a is not None or column_b is not None:
                        table_a = self.control.source_table_a
                        table_b = self.control.source_table_b
                        if column_a is not None and column_b is not None:
                            column_a = table_a.c[column_a]
                            column_b = table_b.c[column_b]
                            column = sa.func.coalesce(column_a, column_b)
                        elif column_a is not None:
                            column = table_a.c[column_a]
                        elif column_b is not None:
                            column = table_b.c[column_b]
                        column = column.label(name) if name else column
                    else:
                        table = self.control.source_table
                        column = table.c[name]
                    columns.append(column)

            process_id = sa.literal(self.control.process_id)
            columns = [*columns, process_id.label('rapo_process_id')]
            select = sa.select(columns)
            select = select.where(sa.literal(1) == sa.literal(0))
            select = select.compile(bind=db.engine, compile_kwargs=db.ckwargs)
            ctas = f'CREATE TABLE {tablename} AS\n{select}'
            index = (f'CREATE INDEX {tablename}_rapo_process_id_ix '
                     f'ON {tablename}(rapo_process_id) COMPRESS')
            compress = (f'ALTER TABLE {tablename} '
                        'MOVE ROW STORE COMPRESS ADVANCED')
            text = db.formatter(ctas, index, compress)
            logger.debug(f'{self.c} Creating table {tablename} '
                         f'with query:\n{text}')
            conn.execute(ctas)
            conn.execute(index)
            conn.execute(compress)
            logger.debug(f'{self.c} {tablename} created')
        table = db.table(tablename)
        return table

    def drop_temporary_tables(self):
        """Clean all temporary tables created during control execution."""
        logger.debug(f'{self.c} Dropping temporary tables...')
        for table in self.control.temp_tables:
            table.drop(db.engine)
        logger.debug(f'{self.c} Temporary tables dropped')
        pass

    def hook(self):
        """Execute database hook procedure."""
        logger.debug(f'{self.c} Executing hook procedure...')
        conn = db.connect()
        process_id = self.control.process_id
        stmt = sa.text(f'begin rapo_control_hook({process_id}); end;')
        conn.execute(stmt)
        logger.debug(f'{self.c} Hook procedure executed')
        pass

    def _fetch_records_to_table(self, select, tablename):
        logger.debug(f'{self.c} Start fetching...')
        conn = db.connect()
        select = select.compile(bind=db.engine, compile_kwargs=db.ckwargs)
        ctas = sa.text(f'CREATE TABLE {tablename} AS\n{select}')
        text = db.formatter(ctas)
        logger.info(f'{self.c} Creating {tablename} with query:\n{text}')
        conn.execute(ctas)
        logger.info(f'{self.c} {tablename} created')
        table = db.table(tablename)
        logger.debug(f'{self.c} Fetching done')
        return table

    def _count_fetched_to_table(self, table):
        logger.debug(f'{self.c} Counting fetched in {table}...')
        conn = db.connect()
        count = sa.select([sa.func.count()]).select_from(table)
        fetched = conn.execute(count).scalar()
        logger.debug(f'{self.c} Fetched in {table} counted')
        return fetched

    pass



"""Setup script for RAPO package."""

import rapo
import setuptools


name = 'rapo'
version = rapo.__version__
author = rapo.__author__
author_email = rapo.__email__
description = rapo.__doc__
long_description = open('README.md', 'r').read()
long_description_content_type = 'text/markdown'
license = rapo.__license__
url = 'https://github.com/t3eHawk/rapo'
install_requires = ['pepperoni', 'psutil',
                    'sqlalchemy', 'cx_oracle', 'sqlparse',
                    'flask', 'flask_httpauth', 'waitress']
packages = setuptools.find_packages()
classifiers = ['Programming Language :: Python :: 3',
               'License :: OSI Approved :: MIT License',
               'Operating System :: OS Independent']


setuptools.setup(name=name,
                 version=version,
                 author=author,
                 author_email=author_email,
                 description=description,
                 long_description=long_description,
                 long_description_content_type=long_description_content_type,
                 license=license,
                 url=url,
                 install_requires=install_requires,
                 packages=packages,
                 classifiers=classifiers)



"""Contains RAPO scheduler interface."""

import argparse
import datetime as dt
import getpass
import json
import os
import platform
import re
import signal
import subprocess as sp
import sys
import threading as th
import time
import queue

import psutil

from ..database import db
from ..logger import logger
from ..reader import reader

from ..main.control import Control


class Scheduler():
    """Represents application scheduler.

    Application scheduler reads configuration from RAPO_CONFIG, schedule
    controls as a virtual jobs and run them when it is necessary in separate
    threads. Number of execution threads is limited by 5.
    Whole scheduling and execution events including errors is being logged
    into simple text files placed to logs folder near by main script with
    Scheduler instance.
    Scheduler timings (current timestamp, delay and waiting) can be seen in
    DEBUG mode.
    Schedule is being updated each 5 minutes from the beginning of the hour.

    Attributes
    ----------
    moment : float or None
        Current scheduler moment - internal timestamp.
    delay : floar or None
        Current scheduler delay - time that is needed to execute internal tasks
        including job scheduling and maintenance events.
    schedule : dict on None
        Dictionary with scheduled jobs where key is a control name and value
        is a control configuration presented as another dictionary.
    queue : queue.Queue
        Queue that consist of jobs that must be executed in FIFO method.
    executors : list
        List of threads that perform job execution.
    server : str or None
        Hostname on which this scheduler is running.
    username : str or None
        OS user that started the scheduler.
    pid : int or None
        OS PID under which this scheduler is running.
    start_date : datetime or None
        Date when scheduler was started.
    end_date : datetime or None
        Date when scheduler was stopped.
    status : str or None
        Current scheduler status.
    """

    def __init__(self):
        self.schedule = None
        self.moment = None
        self.delay = None
        self.queue = queue.Queue()
        self.executors = []
        self.maintenance = th.Event()
        self.maintainer = None

        self.table = db.tables.scheduler
        self.record = reader.read_scheduler_record()
        if self.record and self.record['status'] == 'Y':
            self.server = self.record['server']
            self.username = self.record['username']
            self.pid = int(self.record['pid'])
            self.start_date = self.record['start_date']
            self.stop_date = self.record['stop_date']
            self.status = True if self.record['status'] == 'Y' else False
        else:
            self.server = platform.node()
            self.username = getpass.getuser()
            self.pid = os.getpid()
            self.start_date = None
            self.stop_date = None
            self.status = False

        argv = self._parse_console_arguments()
        if argv:
            action = argv[0]
            if action == 'start':
                self.start()
            elif action == 'stop':
                self.stop()
        else:
            args = self._parse_arguments()
            if args.start is True:
                self._start()
            elif args.stop is True:
                self._stop()
        pass

    @property
    def running(self):
        """Check whether scheduler is running."""
        if self.status is True and self.pid and psutil.pid_exists(self.pid):
            return True
        else:
            return False
        pass

    def start(self):
        """Start scheduler.

        When scheduler is started then normally logs should start to generate
        (in console/file depending on setup).
        RAPO_SCHEDULER will be updated with information about current scheduler
        process including server, username, PID, start date and status.
        """
        return self._create()

    def stop(self):
        """Stop running scheduler.

        Process will be stopped.
        RAPO_SCHEDULER will be updated with stop date and status.
        """
        return self._destroy()

    def read(self):
        """Parse schedule from database table into appropriate structure."""
        return dict(self._sked())

    def _start(self):
        if self.running:
            message = f'scheduler already running at PID {self.pid}'
            raise Exception(message)
        logger.info('Starting scheduler...')
        self.start_date = dt.datetime.now()
        self.status = True
        self._start_signal_handlers()
        self._start_executors()
        self._start_maintainer()
        self._enable()
        logger.info(f'Scheduler started at PID {self.pid}')
        return self._run()

    def _run(self):
        self._synchronize()
        while True:
            self._process()
        pass

    def _stop(self):
        if self.status is True:
            logger.info('Stopping scheduler...')
            self.stop_date = dt.datetime.now()
            self.status = False
            self._disable()
            logger.info(f'Scheduler at PID {self.pid} stopped')
            return self._exit()
        pass

    def _create(self):
        exe = sys.executable
        file = os.path.abspath(sys.argv[0])
        args = '--start'
        settings = {}
        settings['stdout'] = sp.DEVNULL
        settings['stderr'] = sp.DEVNULL
        if sys.platform.startswith('win') is True:
            settings['creationflags'] = sp.CREATE_NO_WINDOW
        command = [exe, file, args]
        proc = sp.Popen(command, **settings)
        return proc

    def _destroy(self):
        if self.status is True:
            self.stop_date = dt.datetime.now()
            self.status = False
            self._disable()
            return self._terminate()
        pass

    def _enable(self):
        conn = db.connect()
        update = self.table.update().values(server=self.server,
                                            username=self.username,
                                            pid=self.pid,
                                            start_date=self.start_date,
                                            stop_date=self.stop_date,
                                            status='Y')
        conn.execute(update)
        pass

    def _disable(self):
        conn = db.connect()
        update = self.table.update().values(stop_date=self.stop_date,
                                            status='N')
        conn.execute(update)
        pass

    def _exit(self):
        return sys.exit()

    def _terminate(self):
        try:
            os.kill(self.pid, signal.SIGTERM)
        except OSError:
            message = f'scheduler at PID {self.pid} was not found'
            raise Warning(message)
        pass

    def _parse_console_arguments(self):
        return [arg for arg in sys.argv[1:] if arg.startswith('-') is False]

    def _parse_arguments(self):
        parser = argparse.ArgumentParser()
        parser.add_argument('--start', action='store_true', required=False)
        parser.add_argument('--stop', action='store_true', required=False)
        args, anons = parser.parse_known_args()
        return args

    def _start_signal_handlers(self):
        logger.debug('Starting signal handlers...')
        signal.signal(signal.SIGINT, lambda signum, frame: self._stop())
        signal.signal(signal.SIGTERM, lambda signum, frame: self._stop())
        logger.debug('Signal handlers started')
        pass

    def _start_executors(self):
        logger.debug('Starting executors...')
        for i in range(5):
            name = f'Thread-Executor-{i}'
            target = self._execute
            thread = th.Thread(name=name, target=target, daemon=True)
            thread.start()
            self.executors.append(thread)
            logger.debug(f'Executor {i} started as {thread.name}')
        logger.debug('All executors started')
        pass

    def _start_maintainer(self):
        logger.debug('Starting maintainer...')
        name = 'Thread-Maintainer'
        target = self._maintain
        thread = th.Thread(name=name, target=target, daemon=True)
        thread.start()
        self.maintainer = thread
        logger.debug(f'Maintainer started as {thread.name}...')
        pass

    def _synchronize(self):
        logger.debug('Time will be synchronized')
        self.moment = time.time()
        logger.debug('Time was synchronized')
        pass

    def _increment(self):
        self.moment += 1
        pass

    def _process(self):
        self._read()
        self._walk()
        self._complete()
        self._next()
        pass

    def _read(self):
        try:
            if self.schedule is None or int(self.moment) % 300 == 0:
                self.schedule = dict(self._sked())
                if self.schedule:
                    logger.debug(f'Schedule: {self.schedule}')
                else:
                    logger.debug('Schedule is empty')
        except Exception:
            logger.error()
        pass

    def _walk(self):
        now = time.localtime(self.moment)
        for name, record in self.schedule.items():
            try:
                if (
                    record['status'] is True
                    and self._check(record['mday'], now.tm_mday) is True
                    and self._check(record['wday'], now.tm_wday+1) is True
                    and self._check(record['hour'], now.tm_hour) is True
                    and self._check(record['min'], now.tm_min) is True
                    and self._check(record['sec'], now.tm_sec) is True
                ):
                    self._register(name, self.moment)
            except Exception:
                logger.error()
        pass

    def _complete(self):
        try:
            if int(self.moment) % 86400 == 0:
                logger.debug('Maintenance triggered')
                self.maintenance.set()
        except Exception:
            logger.error()
        pass

    def _next(self):
        delay = time.time()-self.moment
        wait = 1-delay
        try:
            time.sleep(wait)
        except ValueError:
            logger.warning('TIME IS BROKEN')
            self._synchronize()
        else:
            logger.debug(f'moment={self.moment}, delay={delay}, wait={wait}')
            self._increment()
        pass

    def _sked(self):
        logger.debug('Getting schedule...')
        conn = db.connect()
        table = db.tables.config
        select = table.select()
        result = conn.execute(select)
        for row in result:
            try:
                name = row.control_name
                status = True if row.status == 'Y' else False
                schedule = {} if not row.schedule else json.loads(row.schedule)
                record = {k: v for k, v in schedule.items()
                          if k in ['mday', 'wday', 'hour', 'min', 'sec']}
                record['status'] = status
            except Exception:
                logger.warning()
                continue
            else:
                yield name, record
        logger.debug('Schedule retrieved')
        pass

    def _check(self, unit, now):
        # Check if empty or *.
        if unit is None or re.match(r'^(\*)$', unit) is not None:
            return True
        # Check if unit is lonely digit and equals to now.
        elif re.match(r'^\d+$', unit) is not None:
            unit = int(unit)
            return True if now == unit else False
        # Check if unit is a cycle and integer division with now is true.
        elif re.match(r'^/\d+$', unit) is not None:
            unit = int(re.search(r'\d+', unit).group())
            if unit == 0:
                return False
            return True if now % unit == 0 else False
        # Check if unit is a range and now is in this range.
        elif re.match(r'^\d+-\d+$', unit) is not None:
            unit = [int(i) for i in re.findall(r'\d+', unit)]
            return True if now in range(unit[0], unit[1] + 1) else False
        # Check if unit is a list and now is in this list.
        elif re.match(r'^\d+,\s*\d+.*$', unit):
            unit = [int(i) for i in re.findall(r'\d+', unit)]
            return True if now in unit else False
        # All other cases is not for the now.
        else:
            return False
        pass

    def _register(self, name, moment):
        try:
            logger.info(f'Adding control {name}[{moment}] to queue...')
            self.queue.put((name, moment))
        except Exception:
            logger.error()
        else:
            logger.info(f'Control {name}[{moment}] was added to queue')
        pass

    def _execute(self):
        while True:
            if self.queue.empty() is False:
                name, moment = self.queue.get()
                logger.info(f'Initiating control {name}[{moment}]...')
                try:
                    control = Control(name, timestamp=moment)
                    control.run()
                except Exception:
                    logger.error()
                else:
                    self.queue.task_done()
                    logger.info(f'Control {name}[{moment}] performed')
            time.sleep(1)
        pass

    def _maintain(self):
        while True:
            if self.maintenance.is_set():
                logger.info('Starting maintenance')
                self._clean()
                self.maintenance.clear()
                logger.info('Maintenance performed')
                break
            time.sleep(1)
        pass

    def _clean(self):
        conn = db.connect()
        config = db.tables.config
        select = config.select().order_by(config.c.control_id)
        result = conn.execute(select)
        for row in result:
            control = Control(name=row.control_name)
            control.clean()
        pass

    pass



"""."""

import flask_httpauth

from ...config import config


auth = flask_httpauth.HTTPTokenAuth(scheme='Bearer')
TOKEN = config['API']['token']


@auth.verify_token
def verify_token(token):
    """."""
    if token == TOKEN:
        return token
